/* The U. S. Government Department of Defense wishes to make it known that:
 *
 *
 *				  DISCLAIMER
 *
 *	"Although each program has been tested by its contributor, no
 *	warranty, express or implied, is made by the contributor or the
 *	U. S. Government Department of Defense, as to the accuracy and 
 *	functioning of the program and related program material, nor shall
 *	the fact of distribution constitute any such warranty, and no
 *	responsibility is assumed by the contributor or the U. S. Government
 *	Department of Defense in connection herewith."
 *
 */

/* Subroutines for MIME message packer/unpacker
   Written by M. Byrne (mbyrne@accurate.com)
   */

#include "sgmlincl.h"         /* #INCLUDE statements for SGML parser. */
#include "pack.h"
#include "appl.h"
#include <sys/time.h>
#include <errno.h>

#define M_OPENC	1			/* can't open configuration file*/
#define M_FSEEK	2
#define M_OPENMSGW	3		/* can't open message file	*/
#define M_OPENMSGA	4

#define BPERLIN	76

extern char flushed;
extern char parsflushed;
extern char *savrbufs;
extern FILE *debugfp;

#ifndef lint
static char sccsid[] = "@(#) ID: pack.c  V 1.9, 9/23/94 at 11:18:23 mbyrne@accurate.com";
#endif

struct ENTITYLIST entitylist[MAXFNAMES];
struct NOTATLIST notatlist[MAXEXTIDS];
struct CONTENTYPE contentype[MAXEXTIDS];
struct IDLIST idlist[MAXFNAMES];
struct NEST nest[QENTLVL];

int saviorc;
int thispid;
char *home;
char packdef;
char havSDIF;
char doneabove;
char configfile[FNAMELEN];
char msgname[FNAMELEN];
char msgheader[FNAMELEN];
char boundary[100];
char addtoboundary[50];
char descriptor[100];
char ascidocnum[4];

char namebuf[50];
char *name;
int namelen;

char syscmd[100];

static char desc_default[20];		/* Content-descriptor default */
static char currentype;			/* related A, main doc. M, related B */
static char lastype;
static char related;
static int  currentdocnum;
static VOID do_merror P((int));
static char *mimerrlist[] = {
     0,
     "Can't open configuration file",		/* M_OPENC	*/
     "fseek failed on configuration file",	/* M_FSEEK	*/
     "Can't open message file",			/* M_OPENMSGW	*/
     "Can't append to message file",		/* M_OPENMSGA	*/
};
static char *prog;		/* Program name (for error messages)	*/

FILE *mm;			/* Output message			*/
FILE *mmctypes;
FILE *hdrmm;			/* Headers for output message		*/

int filecount;

char *cont_buf;
int countypes=0;
int entitycount=0;
int notatcount=0;

int
init_contentype() 
{
	/* Called by packer/unpacker to initialize the contentype array
	   */
	long lret,offset;
	int i,j,ret;
	char c,inquotes;

	flushed = NO;		/* init the flushed flag	*/
	parsflushed = NO;
	doneabove = NO;

	/* Read the configuration file and generate
	   a table of pointers to each "sgmls-" entry	*/
	prog = "pack";

     	errno = 0;
     	mmctypes = fopen(configfile, "r");
     	if (!mmctypes) {
		do_merror(M_OPENC);
		return(FAILURE);
	}

	errno = 0;
	offset = 0;

	/* seek to SEEK_END (whence = 2) of file to find the file size
	   and allocate space enough to hold it			*/
	if ((lret = fseek(mmctypes,offset,2)) == -1) {
		fprintf(stderr,"pack: init_contentype: ERROR: errno from fseek %ld\n",errno);
		do_merror(M_FSEEK);
		return(FAILURE);
	}
	lret = ftell(mmctypes);

        cont_buf = (char *)xmalloc(lret+1);   /* +1 for null at begin of buffer */
	rewind(mmctypes);

	inquotes = NO;
	i = j = 0;
	cont_buf[0] = '\0';

	/* Copy entire configuration file into memory.
	   Set pointers as we encounter "sgmls-" entries */
	while ((c = getc(mmctypes)) != EOF) {
		if (c == '\"') {
			if (inquotes == YES ) {
				inquotes = NO;
				cont_buf[i++] = '\0';
			} else {
				inquotes = YES;
				cont_buf[i++] = c;
			}
			continue;
		}
		if (c == ':') {
			cont_buf[i++] = '\0';
			continue;
		}
		/* Bump over white space.  Set it to NULL	*/
		if (((c == ' ') || (c == '	')) && (inquotes == NO)) {
			cont_buf[i++] = '\0';
			continue;
		}
		if (c == '\n') {
			cont_buf[i++] = '\0';	/* replace newline with NULL */
			contentype[countypes].extype = &cont_buf[j];
			if (strncmp(contentype[countypes].extype,"sgmls-",6) == 0) {
				/* move extype ptr past sgmls- string  */
			   	if (contentype[countypes].extype[6] == '\"') {
					contentype[countypes].extype += 7*(sizeof(char));
					j += 7*(sizeof(char));
			   	}else {
					contentype[countypes].extype += 6*(sizeof(char));
					j += 6*(sizeof(char));
			   	}
				ret = strlen(contentype[countypes].extype);
				contentype[countypes].extype[ret] = '\0';
				j += ret;
				while(cont_buf[j] == '\0') {
					j++;
				}
				contentype[countypes].mimetype = &cont_buf[j];
				countypes++;
			}
			j = i;
			continue;
		}
		if (cont_buf[i-1] == '\0') {
		}
		cont_buf[i++] = c;
	}
	/* print the contentype table		*/
     	if (debugfp) {
		fprintf(debugfp,"\ninit_contentype: countypes = %d, contentype table follows:\n",countypes);
		for (i=0; i<countypes; i++) {
			fprintf(debugfp,"i = %d, extid:	%s,	mimetype: %s\n",i,
				contentype[i].extype,
				contentype[i].mimetype);
		}
		fprintf(debugfp,"\n");
		fflush(debugfp);
     	}
	return(SUCCESS);
}

doSDIF(argc,argv,argvindex)
int argc;
char **argv;
int argvindex;
{
	int i,n,len,ret;
	char dodoc;
	char *docfiles[10];

     	if (debugfp) {
		fprintf(debugfp,"\npack: doSDIF: Enter. argc %d, argvindex %d\n",argc,argvindex);
		for (i=0; i<argc; i++) {
			fprintf(debugfp,"doSDIF: argv[%d] %s\n",i,argv[i]);
		}
		fflush(debugfp);
     	}

	if((ret = sanity(argc,argv,argvindex)) == FAILURE) {
	}

	if(descriptor[0] != '\0')
		fprintf(hdrmm, "Content-descriptor: %s\n",descriptor);
	fprintf(hdrmm, "Content-Type: Multipart/SDIF");
	fprintf(hdrmm, ";\n   boundary=\"%s\"",boundary);
	fprintf(hdrmm, "\n\n--%s\n",boundary);
	fflush(hdrmm);
	strcpy(descriptor,desc_default);	/* reinit the descriptor */

	n = 0;
	dodoc = NO;
	currentdocnum = 0;
	currentype='Z';
	lastype='Z';
	related = 'A';

     	for (i=argvindex; i<argc; i++) {
		len = strlen(argv[i]);
		/*
		if (debugfp) {
			fprintf(debugfp,"\ndoSDIF: argv[%d] %s, len %d, n %d\n",i,argv[i],len,n);
			fprintf(debugfp,"doSDIF: look at %s \n",argv[i]);
			fflush(debugfp);
		}
		*/
		if ((strncmp(argv[i],"-R",2)==0) ||
			(strncmp(argv[i],"-M",2)==0)) 
		{
		   if(dodoc == YES) {
			if ((ret = newdoc_start(currentype)) == FAILURE ) {
       				fprintf(stderr,"%s: ERROR. newdoc_start failed\n",prog);
       				fflush(stderr);
       				return(FAILURE);
			}

			if (sgmlsdoc(make_docent(n,docfiles)))
				return(E_DOC);

			process_document(0);

			if ((ret = output_mimemsg()) != SUCCESS) {
				fprintf(stderr,"ERROR. %s failed in output_mimemsg\n",prog);
				fflush(stderr);
				return(FAILURE);
			}

			/* output_mimemsg left all info in hdrmm
			   and files closed			*/
			if ((ret = start_output("a")) == FAILURE) {
				fprintf(stderr,"%s: ERROR. start_output failed\n",prog);
				fflush(stderr);
				return(FAILURE);
			}
			dodoc = NO;
			n = 0;
		   }
		   currentype = argv[i][1];
		   /*
		   if (debugfp) {
		   	fprintf(debugfp,"doSDIF: set currentype to %c\n",currentype);
		   	fflush(debugfp);
		   }
		   */
		}else {
		   if (strncmp(argv[i],"-D",2)==0) {
			if (len == 2) {
				i++;
				strcpy(descriptor,argv[i]);
			}else{
				strcpy(descriptor,&argv[i][2]);
			}
		   }else {
			docfiles[n] = argv[i];
			dodoc = YES;
			n++;
		   }
		}
     	}
	/* Process last document			*/
	if ((ret = newdoc_start(currentype)) == FAILURE ) {
	   	fprintf(stderr,"%s: ERROR. newdoc_start failed\n",prog);
	       	fflush(stderr);
	       	return(FAILURE);
	}

	if (sgmlsdoc(make_docent(n,docfiles)))
		return(E_DOC);

	process_document(0);

	if ((ret = output_mimemsg()) != SUCCESS) {
		fprintf(stderr,"ERROR. %s failed in output_mimemsg\n",prog);
		fflush(stderr);
		return(FAILURE);
	}

	if ((ret = start_output("a")) == FAILURE) {
		fprintf(stderr,"%s: ERROR. start_output failed\n",prog);
		fflush(stderr);
		return(FAILURE);
	}

	if (currentype == 'R') {
		fprintf(hdrmm,"--%s%s--\n",boundary,addtoboundary);
	}
	fprintf(hdrmm, "--%s--\n",boundary);
	fclose(hdrmm);
	fclose(mm);
	return(SUCCESS);
}

donoSDIFpart(argc,argv,argvindex)
int argc;
char **argv;
int argvindex;
{
	int i,ret;
     	if (debugfp) {
		fprintf(debugfp,"\npack: donoSDIFpart: Enter. argc %d\n",argc);
		for (i=0; i<argc; i++) {
			fprintf(debugfp,"donoSDIFpart: argv[%d] %s\n",i,argv[i]);
		}
		fflush(debugfp);
     	}
	if (descriptor[0] != '\0')
		fprintf(hdrmm, "Content-descriptor: %s\n",descriptor);
	fprintf(hdrmm, "Content-Type: Multipart/SGML;\n");
	fprintf(hdrmm, "   boundary=\"%s\"",boundary);
        if (sgmlsdoc(make_docent(argc - argvindex, argv + argvindex)))
	 	return(E_DOC);

	process_document(0);

	if ((ret = output_mimemsg()) != SUCCESS) {
		fprintf(stderr,"ERROR. %s failed in output_mimemsg\n",prog);
		fflush(stderr);
		return(FAILURE);
	}
} 

start_output(mode)
char * mode;
{
	if (debugfp) {
		fprintf(debugfp,"pack: start_output: Enter. mode %c, currentype %c, lastype %c\n",*mode,currentype,lastype);
		fflush(debugfp);
	}

	/* Start the mime output message output file			*/
     	errno = 0;
	mm = fopen(msgname,"w");
     	if (!mm)  {
		do_merror(M_OPENMSGW);
		return(FAILURE);
	}
	sprintf(msgheader,"hdr%d",thispid);
     	hdrmm = fopen(msgheader, mode);
     	if (!hdrmm)  {
		do_merror(M_OPENMSGW);
		return(FAILURE);
	}

	if (debugfp) {
		fprintf(debugfp,"pack: start_output, opened %s mode %c, and %s\n",msgheader,*mode,msgname);
	}
	nest[0].fp = mm;
	strcpy(nest[0].tempfile,msgname);
	return(SUCCESS);
}

newdoc_start(doctype)
char doctype;			/* can be only R or M	*/
{
	/* This routine handles the breaks in major parts 
	   (related A, main doc, related B)
	   */

	currentdocnum++;
	addtoboundary[0] = '\0';

	if (debugfp) {
		fprintf(debugfp,"newdoc_start. Enter with doctype %c, lastype %c, currentype %c, related %c, currentdocnum %d\n",doctype,lastype,currentype,related,currentdocnum);
		fflush(debugfp);
	}

	ascidocnum[0] = '-';
	longitoa((long)currentdocnum,&ascidocnum[1]);

	if (doctype != lastype) {
		if (doctype == 'R') {
			if (related == 'B') {
				/* Bound main doc part		*/
				fprintf(hdrmm,"\n\n--%s\n",boundary);
			}
			fprintf(hdrmm, "Content-type: multipart/Mixed");
			sprintf(addtoboundary,"-RELATED-DOC-%c",related);
			fprintf(hdrmm, ";\n   boundary=\"%s%s\"",boundary,addtoboundary);
			fprintf(hdrmm,"\n\n--%s%s\n",boundary,addtoboundary);
		}else{
			if (lastype == 'R') {
				/* Bound related A part		*/
				fprintf(hdrmm,"\n--%s-RELATED-DOC-%c--\n",boundary,related);
				fprintf(hdrmm,"\n\n--%s\n",boundary);
			}
			sprintf(addtoboundary,"-MAIN-DOC");
			ascidocnum[0] = '\0';
			related = 'B';
		}
		lastype = currentype;
	} else {
		if(doctype != 'R') {
			fprintf(stderr,"ERROR. %c on command line after %c\n",doctype,lastype);
			fflush(stderr);
			return(FAILURE);
		}
		sprintf(addtoboundary,"-RELATED-DOC-%c",related);
		fprintf(hdrmm,"\n--%s%s\n",boundary,addtoboundary);
	}

	if (descriptor[0] != '\0')
		fprintf(hdrmm, "Content-descriptor: %s\n",descriptor);
	fprintf(hdrmm, "Content-type: multipart/SGML;\n");
	fprintf(hdrmm, "   boundary=\"%s%s%s\"",boundary,addtoboundary,ascidocnum);
	fflush(hdrmm);

	/* Leave the files open for fileread() in sgml2.c to copy the
	   input into it  and for parsepro in pars1.c to add headers
	   as appropriate						*/
	return(SUCCESS);
}

int
newMIMEpart(outfp,type,id)
FILE *outfp;
char *type;
char *id;
{
	int i;
	char buf[BUFSIZ];

     	if (debugfp) {
		fprintf(debugfp,"\npack: newMIMEpart: Enter. outfp %ld, type %s, id %s\n",outfp,type,id);
		fflush(debugfp);
     	}

	if (packdef == PACK) {
		fprintf(outfp, "\n\n--%s%s%s\n",boundary,addtoboundary,ascidocnum);
		fprintf(outfp, "Content-ID: \"%s\"\n",id);
		fprintf(outfp, "Content-Type: %s\n",type);
	} else {	/* UNPACK	*/
		/* Close the file for the previous body part.
		   Open the next.  Have to find its name in idlist.   */
		sprintf(buf,"%s",id);
		for (i=0; i<filecount; i++) {
			if (strcmp(buf,idlist[i].filename) == 0) {
				fclose(outfp);
     				if (debugfp) {
				   fprintf(debugfp,"pack: newMIMEpart: Close %ld, Open %s for write\n",
					outfp,idlist[i].tempfile);
				   fflush(debugfp);
     				}
				errno = 0;
				outfp = fopen(idlist[i].tempfile,"w");
				if (!outfp) {
					fprintf(stderr,"ERROR: can't open file %s, errno %ld\n",idlist[i].tempfile,errno);
					fflush(stderr);
					return(FAILURE);
				}
				nest[0].fp = outfp;
				strcpy(nest[0].tempfile,idlist[i].tempfile);
				idlist[i].pid = thispid;
				break;
			}
		}
	}
	return(SUCCESS);
}

int
output_mimemsg()
{
	/* Called at end of processing a document. It finishes up
	    the output file(s)				*/
	int i,j,k,ret;
	char restoretemp;
     	if (debugfp) {
		fprintf(debugfp,"\noutput_mimemsg: Enter. entitycount = %d, notatcount %d, countypes %d\n",entitycount,notatcount,countypes);
		fprintf(debugfp,"\noutput_mimemsg: notatlist table follows:\n");
		for (i=0; i<notatcount; i++) {
			fprintf(debugfp,"i = %d, extid:	%s,	mimeid: %s\n",i,
				notatlist[i].extid,
				notatlist[i].mimeid);
		}
		fflush(debugfp);
     	}
	if (nest[0].fp != 0) {
		fclose(nest[0].fp);
	}

     	if (debugfp) {
		for (i=0; i<entitycount; i++) {
			fprintf(debugfp,"output_mimemsg: i %d, oldname %s \n",
				i,entitylist[i].oldname);
			fprintf(debugfp,"output_mimemsg: i %d, newname %s \n",
				i,entitylist[i].newname);
			fprintf(debugfp,"output_mimemsg: i %d, tempfile %s \n",
				i,entitylist[i].tempfile);
			fprintf(debugfp,"output_mimemsg: i %d, extid %s \n",
				i,entitylist[i].extid);
			fprintf(debugfp,"output_mimemsg: i %d, docnum %d \n",
				i,entitylist[i].docnum);
			fprintf(debugfp,"output_mimemsg: i %d, alreadyout %d\n",
				i,entitylist[i].alreadyout);
		}
		fflush(debugfp);
     	}

	if (packdef == PACK) {
	   /* Input file has already been written into the message.
	      Concatenate it with the headers then
	      throw in the files that were referenced internally */

	   prog = "pack";
	   fclose(hdrmm);

	   for (i=0; i<entitycount; i++) {
	     if (entitylist[i].docnum == currentdocnum) {
     		errno = 0;
     		mm = fopen(msgname, "a");
     		if (!mm)  {
			do_merror(M_OPENMSGA);
			return(FAILURE);
		}

		/* extra newline is for receiver's mhn-store	*/
		fprintf(mm, "\n\n--%s%s%s\n",boundary,addtoboundary,ascidocnum);

	     if(entitylist[i].alreadyout == YES) {
			fprintf(mm, "Content-Type: Application/SDIF;\n	Cross-Reference=\"%s\"\n",entitylist[i].newname);
			fclose(mm);
			continue;
	     } 

		k = 0;
		if (entitylist[i].extid[0] == '\0') {
			fprintf(mm, "Content-Type: Application/sgml\n");
		} else {
		   for (j=0; j<notatcount; j++) {
			/* match ENTITY extid to NOTATION extid */
			ret = strcmp(entitylist[i].extid,notatlist[j].extid);
			if (ret == 0) {
				/* match NOTATION id to .mh_contypes entry */
				for (k=0; k<countypes; k++) {
					ret=strcmp(notatlist[j].mimeid,contentype[k].extype);
					if (ret == 0) {
					   break;
					}
				}
				break;
			}
		   }
		   if (countypes == 0) {
			/* no entries in config file - use x_Unknown	*/
			fprintf(mm, "Content-Type: x_Unknown\n");
		   }else {
		   	if (k == countypes) k--;
			fprintf(mm, "Content-Type: %s\n",contentype[k].mimetype);
		   }
		}	/* end of else	*/

		fprintf(mm, "Content-ID: \"%s\"\n",entitylist[i].newname);

		restoretemp = NO;
		entitylist[i].alreadyout = YES;
		if (entitylist[i].tempfile[0] == '\0') {
			strcpy(entitylist[i].tempfile,entitylist[i].oldname);
			restoretemp = YES;
		} 
		if ((ret = checkfile(entitylist[i].tempfile)) != SUCCESS)
		{
			fprintf(stderr,"ERROR: output_mime_msg got FAILURE from checkfile %s\n", entitylist[i].tempfile);
			return(ret);
		}
		sprintf(syscmd,"cat %s >> %s",entitylist[i].tempfile,msgname);

		/* add message body, beginning with an empty line */
		fprintf(mm, "\n");
		fclose(mm);
		system(syscmd);
     		if (debugfp) {
			fprintf(debugfp,"syscmd -> %s\n",syscmd);
			fflush(debugfp);
     		}
		if (restoretemp == YES) {
			entitylist[i].tempfile[0] = '\0';
			restoretemp = NO;
		}
	     }	/* not part of current document	*/
	   }

		/* Add closing part boundary		*/
     		errno = 0;
     		mm = fopen(msgname, "a");
     		if (!mm) {
			fprintf(stderr,"pack: ERROR. output_mime.msg fopen failed for file %s, errno %d\n",msgname,errno);
			do_merror(M_OPENMSGA);
			return(FAILURE);
		}
		fprintf(mm, "\n--%s%s%s--\n",boundary,addtoboundary,ascidocnum);
		fclose(mm);
	        sprintf(syscmd,"cat %s >>%s",msgname,msgheader);
	        system(syscmd);

		reinit();
		return(SUCCESS);
	} else {	/* UNPACK	*/
     		if (debugfp) {
			fprintf(debugfp,"\npack: output_mime.msg: filecount %d, thispid %d, IDLIST table follows:\n",filecount,thispid);
			for (i=0; i<filecount; i++) {
				fprintf(debugfp,"%d. filename: %s,\n",
					i,idlist[i].filename);
				fprintf(debugfp,"%d. tempfile: %s\n",
					i,idlist[i].tempfile);
				fprintf(debugfp,"%d. pid: %ld\n",
					i,idlist[i].pid);
			}
     		}

		for (i=0; i<filecount; i++) {
			if ((idlist[i].tempfile[0] != '\0') && (idlist[i].pid == thispid)) {
				sprintf(syscmd,"mv %s %s",
					idlist[i].tempfile,
					idlist[i].filename);
     				if (debugfp) {
					fprintf(debugfp,"output_mime_msg: syscmd %s\n",syscmd);
     				}
				system(syscmd);
			}
		}
	}
	return(SUCCESS);
}

static
VOID do_merror(errnum)
int errnum;
{
     fprintf(stderr, "%s: %s\n", prog,mimerrlist[errnum]);
     fflush(stderr);
}


entitySYS(estore,ename)
int estore;
UNCH *ename;		/* entity name			*/
{
	/* Called by both the packer and the unpacker.
	   The packer builds a filename epoctime.host.domain
	   The unpacker uses the name that was passed in the input table
	   */
	int i,j;
	int index;

	index = entitycount;

	/* Check to see if it's already on the list.  If so don't duplicate */
	for (i=0; i<entitycount; i++) {
		if (strcmp(ename,entitylist[i].oldname) == 0)
		{
			entitylist[i].docnum = currentdocnum;
			index = i;
		}
	}

	strcpy (&entitylist[index].oldname[0],ename);

	/* find start of file name			*/
	/* i tracks rbufs,  j tracks savrbufs		*/
	i = j = 0;
	while(rbufs[i] != '\"') {
		savrbufs[j] = rbufs[i];
		i++; j++;
	}
	savrbufs[j++] = '\"';

	if (index == entitycount) {
		buildnewname(estore,ename);
	}
	strcpy(&savrbufs[j],entitylist[index].newname);
	j += strlen(entitylist[index].newname);

	savrbufs[j++] = '\"';
		
	/* find second quote after old name	*/
	i++;
	while (rbufs[i++] != '\"') ;

	/* copy the rest of the line		*/
	while (i <= saviorc) {
		savrbufs[j] = rbufs[i];
		i++; j++;
	}
	saviorc = j-1;		/* adjust the line length for output */
	savrbufs[j] = '\0';
}

/* from page 60 of "The C Programming Language", Kernighan and Ritchie
*/
longitoa(n,s)
char s[];
long n;
{
	int i;
	i = 0;
	do {
		s[i++] = n % 10 + '0';
	} while ((n /= 10) >0);
	s[i] = '\0';
	reverse(s);
}
reverse(s)
char s[];
{
	int c, i, j;
	for (i=0, j=strlen(s)-1; i<j; i++, j--) {
		c = s[i];
		s[i] =s[j];
		s[j] = c;
	}
}

clipitoa(n,s)
char s[];
long n;
{
	int i;
	i = 0;
	do {
		s[i++] = n % 10 + '0';
	} while ((n /= 10) >0);
	s[i] = '\0';
	reverse(s);

	/* clip off trailing zeros 		*/
	for (i=strlen(s)-1; i>0; i--) {
		if (s[i] == '0') {
			s[i] = '\0';
		} else {
			break;
		}
	}
}

int
readtable(filename,argc,argv)
char ** filename;
int argc;
char **argv;
{
/* Read the table produced by the receiving mail agent.  Each line contains
   a storage file name and the Content-ID of each body part.
   These two pieces of information are separated by a single TAB.

   This subroutine fills IDLIST structure
   */
	FILE *tp;
	char buf[BUFSIZ];
	char c;
	int mid,i,j,curfile;

	if (debugfp) {
		fprintf(debugfp,"pack: readtable enter with %s\n",filename);
		fprintf(debugfp,"pack: readtable: argc %d\n",argc);
        	for (i = 0; i < argc; i++) {
     			fprintf(debugfp,"pack: readtable: i %d, file %s\n",i,argv[i]);
		}
		fflush(debugfp);
	}
	errno = 0;
	tp = fopen(filename,"r");
	if (!tp) {
		fprintf(stderr,"ERROR: can't open %s, errno %ld\n",filename,errno);
		return (FAILURE);
	}

	j = 0;
	curfile = 0;

	while ((c=fgetc(tp)) != EOF) {
		buf[j] = c;
		if (c =='\t') {
			buf[j] = '\0';
			mid = ++j;
		} else {
			if (c == '\n') {
				buf[j] = '\0';
				strcpy(idlist[curfile].filename,buf);
				strcpy(idlist[curfile].contentid,&buf[mid]);
				sprintf(idlist[curfile].tempfile,"temp.%d.%d",curfile,thispid);
				idlist[curfile].pid = 0; /* filled in later when
						            actual file is 
							    opened	*/
				curfile++;
				j = 0;
			} else {
				j++;
			}
		}
	}
	fclose(tp);
	filecount = curfile;


	if (debugfp) {
		fprintf(debugfp,"\npack: readtable: filecount %d, IDLIST table follows:\n",filecount);
		fflush(debugfp);
		for (i=0; i<filecount; i++) {
			fprintf(debugfp,"i %d, filename %s\n",
				i,idlist[i].filename);
			fprintf(debugfp,"i %d, contentid %s\n",i,idlist[i].contentid);
			fprintf(debugfp,"i %d, tempfile %s\n\n",i,idlist[i].tempfile);
		}
		fflush(debugfp);
	}

	/* Translate command line argvs into actual file names.
	   */
	for (j=0; j<argc; j++) {
	   for (i=0; i<filecount; i++) {
		sprintf(buf,"\"%s\"%c",argv[j],NULL);
		if (strcmp(buf,idlist[i].contentid) == 0)
		{
			argv[j] = &idlist[i].filename[0];
			if(debugfp) {
				fprintf(debugfp,"replace argv to point to %s \n",
				   argv[j]);
			}
			if (j==0) {
				strcpy(nest[0].tempfile, idlist[i].tempfile);
			}
			break;
		}
	   }
	}

	mm = fopen(nest[0].tempfile, "w");
     	if (!mm)  {
		do_merror(M_OPENMSGW);
		return(FAILURE);
	}
	nest[0].fp = mm;
     	if (debugfp) {
		fprintf(debugfp,"pack: readtable: opened nest[0].tempfile = %s\n",nest[0].tempfile);
		fflush(debugfp);
     	}

	return(SUCCESS);
}

static int checkfile(name) 
char *name;
{
	int ret;
	FILE *extfp,*tmpfp;
	char tmpfile[FNAMELEN];
	char c;

	errno = 0;
	if ((extfp = fopen(name,"r")) == 0) {
		fprintf(stderr,"ERROR: checkfile can't open %s, errno %d\n",name,errno);
		return (FAILURE);
	}

	while ((c = fgetc(extfp)) != EOF) {
		ret = isascii(c);
		if (ret == 0) {
			fprintf(mm, "Content-transfer-encoding: Base64\n");
			/* Not ascii.  Need to encode the file	*/
			rewind(extfp);
			sprintf(tmpfile,"tmp.%d",thispid);
			errno = 0;
			tmpfp = fopen(tmpfile,"w");
			if (!tmpfp) {
				fprintf(stderr,"ERROR: checkfile can't open tmpfile %s, errno %d\n",tmpfile,errno);
				fflush(stderr);
				return(FAILURE);
			}
			writeBase64(extfp,tmpfp);
			strcpy(name,tmpfile);	/* overwrite file name */
			fclose(tmpfp);
			fclose(extfp);
			return(SUCCESS);
		}
	}
	return(SUCCESS);
}


/* Following taken from Marshall T. Rose, mhn.c		*/

static char nib2b64[0x40+1] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static int  writeBase64 (in, out)
FILE   *in,
       *out;
{
    int	    cc,
	    n;
    char    inbuf[3];

    n = BPERLIN;
    while ((cc = fread (inbuf, sizeof *inbuf, sizeof inbuf, in)) > 0) {
	unsigned long bits;
	register char *bp;
	char    outbuf[4];

	if (cc < sizeof inbuf) {
	    inbuf[2] = 0;
	    if (cc < sizeof inbuf - 1)
		inbuf[1] = 0;
	}
	bits = (inbuf[0] & 0xff) << 16;
	bits |= (inbuf[1] & 0xff) << 8;
	bits |= inbuf[2] & 0xff;

	for (bp = outbuf + sizeof outbuf; bp > outbuf; bits >>= 6)
	    *--bp = nib2b64[bits & 0x3f];
	if (cc < sizeof inbuf) {
	    outbuf[3] = '=';
	    if (cc < sizeof inbuf - 1)
		outbuf[2] = '=';
	}

	(void) fwrite (outbuf, sizeof *outbuf, sizeof outbuf, out);

	if (cc < sizeof inbuf) {
	    (void) putc ('\n', out);
	    return SUCCESS;
	}

	if (--n <= 0) {
	    n = BPERLIN;
	    (void) putc ('\n', out);
	}
    }
    if (n != BPERLIN)
	(void) putc ('\n', out);

    return SUCCESS;
}



sanity(argc,argv,argvindex)
int argc;
char **argv;
int argvindex;
{
	/* Check command line for:
		All files exist
		One and only one -M
		Logical sequence
			[-R [-Dtext] related A files]
			-M [-Dtext] doc files 
			[-R [-Dtext] related B files]
		-O cannot appear with -U
		-P cannot appear with -U
		-N cannot appear with -U
		-R cannot appear with -U
		-B cannot appear with -U
		-N calls for -M
	*/
	return;
}

buildnewname(estore,ename)
int estore;
UNCH *ename;		/* entity name			*/
{
	int k,kk,ret;
	char buf[BUFSIZ];
	struct timeval tval;
	struct timeval *tp;
	struct timezone *tzp;
	struct timezone tzone;

	/* build a new file name			*/
	if (packdef == PACK) {
		kk = 0;
	
		entitylist[entitycount].docnum = currentdocnum;
		/* pick up entity name				*/
		if (debugfp) {
			fprintf(debugfp,"pack: buildnewname: nmbuf[1] %c, pero is %c, estore %d\n",nmbuf[1],lex.d.pero,estore);
			fflush(debugfp);
		}
		if (nmbuf[1] == lex.d.pero) {
			strcpy(entitylist[entitycount].newname,nmbuf+2);
		} else {
			if (estore == ESD) {
				entitylist[entitycount].extid[0] = '\0';
			} else {
				strcpy(entitylist[entitycount].newname,nmbuf+1);
			}
		}
		k = strlen(entitylist[entitycount].newname);
		kk += k;
	
	
		/* get time in seconds since Jan. 1, 1970?	*/
		tp = &tval;
		tzp = &tzone;
	
		if ((ret = gettimeofday(tp,tzp)) == 0) {
			if (kk >0) {
				entitylist[entitycount].newname[kk++] = '.';
			}
			longitoa(tp->tv_sec,&entitylist[entitycount].newname[kk]);
			k = strlen(&entitylist[entitycount].newname[kk]);
			kk += k;
	
			entitylist[entitycount].newname[kk++] = '.';
			clipitoa(tp->tv_usec,&entitylist[entitycount].newname[kk]);
			k = strlen(&entitylist[entitycount].newname[kk]);
			kk += k;
		} else {
			if (debugfp) {
			fprintf(debugfp,"ret from gettimeofday %d, errno %ld\n",ret,errno);
			}
		}
	
		namelen = 50;
		name = namebuf;
		if ((ret = gethostname(name,namelen)) == 0) {
			if (namelen >0) {
				entitylist[entitycount].newname[kk++] = '\@';
				strcpy (&entitylist[entitycount].newname[kk],name);
				kk += strlen(name);
			}
		} else {
			fprintf(stderr,"pack: ERROR. bad ret from gethostname, errno %ld\n",errno);
		}
	
		namelen = 50;
		name = namebuf;
		if ((ret=getdomainname(name,namelen)) == 0) {
			if (namelen >0) {
				entitylist[entitycount].newname[kk++] = '.';
				strcpy (&entitylist[entitycount].newname[kk],name);
				kk += strlen(name);
			}
		} else {
			fprintf(stderr,"pack: ERROR. bad ret from getdomainname, errno %ld\n",errno);
		}
	} else {	/* UNPACKer finds new name in IDLIST		*/
		sprintf(buf,"\"%s\"%c",entitylist[entitycount].oldname,NULL);
		for (k=0; k<filecount; k++) {
			if (strcmp(buf,idlist[k].contentid) == 0) {
				strcpy(entitylist[entitycount].newname,idlist[k].filename);
				break;
			}
		}
		if (debugfp) {
			fprintf(debugfp,"buildnewname: oldname %s\n",
				entitylist[entitycount].oldname);
			fprintf(debugfp,"buildnewname: newname %s\n",
				entitylist[entitycount].newname);
			fflush(debugfp);
		}
	}

	entitylist[entitycount].tempfile[0] = '\0'; 
	entitycount++;
}

/* generate a Content-ID for part of the main document */
genid(buf)
char *buf;
{
	int k,ret;
	struct timeval tval;
	struct timeval *tp;
	struct timezone *tzp;
	struct timezone tzone;

	/* get time in seconds since Jan. 1, 1970?	*/
	tp = &tval;
	tzp = &tzone;
	k = 0;

	errno = 0;
	if ((ret = gettimeofday(tp,tzp)) == 0) {
		longitoa(tp->tv_sec,buf);
		k = strlen(buf);

		buf[k++] = '.';
		clipitoa(tp->tv_usec,&buf[k]);
		k = strlen(buf);
	}else {
		fprintf(stderr,"pack: genid: ERROR from gettimeofday, errno %ld\n",errno);
	}

	namelen = 50;
	name = namebuf;
	errno = 0;
	if ((ret = gethostname(name,namelen)) == 0) {
		if (namelen >0) {
			buf[k++] = '\@';
			strcpy (&buf[k],name);
			k = strlen(buf);
		}
	} else {
		fprintf(stderr,"pack: genid: ERROR from gethostname, errno %ld\n",errno);
	}

	namelen = 50;
	name = namebuf;
	errno = 0;
	if ((ret=getdomainname(name,namelen)) == 0) {
		if (namelen >0) {
			buf[k++] = '.';
			strcpy (&buf[k],name);
		}
	} else {
		fprintf(stderr,"pack: genid: ERROR from getdomainname, errno %ld\n",errno);
	}
	fflush(stderr);
	return;
}

int
setpackdefaults()
{
     /* Set defaults						*/
     packdef = SGMLS;			/* program invoked 	*/
     debugfp = 0;
     strcpy(msgname,MSGNAME);		/* out msg name		*/
     strcpy(boundary,BOUNDARY);		/* part boundary	*/
     strcpy(desc_default,DESCRIPTOR);
     strcpy(descriptor,desc_default);	/* Content-Descriptor   */
     havSDIF = NO;
     errno = 0;
     if ((home = getenv("HOME")) == (char *)NULL) {
	  fprintf(stderr,"pack: ERROR: Can't find HOME in getenv(), errno %ld\n",errno);
	  fflush(stderr);
	  return(FAILURE);
     }
     strcpy(configfile,home);
     strcat(configfile,"/");
     strcat(configfile,CONFIGFILE);	/* configuration file	*/
     thispid = getpid();
     return(SUCCESS);
}

reinit()
{
badresw = 0;              /* 1=REF_ out of context; 0=valid. */
charmode = 0;             /* >0=in #CHARS; 0=not. */
conactsw = 0;             /* 1=return saved content action 0=get new one.*/
conrefsw = 0;             /* 1=content reference att specified; 0=no. */
contersv = 0;             /* Save contersw while processing pending REF. */
contersw = 0;             /* 1=element or #CHARS out of context; 0=valid. */
datarc = 0;               /* Return code for data: DAF_ or REF_. */
delmscsw = 0;             /* 1=DELMSC must be read on return to es==0. */
didreq = 0;               /* 1=required implied tag processed; 0=no. */
dostag = 0;               /* 1=retry newetd instead of parsing; 0=parse. */
dtdsw = 0;                /* DOCTYPE declaration found: 1=yes; 0=no. */
entdatsw = 0;             /* 2=CDATA entity; 4=SDATA; 8=NDATA; 0=none. */
entpisw = 0;              /* 4=PI entity occurred; 0=not. */
eodsw = 0;                /* 1=eod found in error; 0=not yet. */
eofsw = 0;                /* 1=eof found in body of document; 0=not yet. */
es = -1;                  /* Index of current source in stack. */
etagimct = 0;             /* Implicitly ended elements left on stack. */
etagimsw = 0;             /* 1=end-tag implied by other end-tag; 0=not. */
etagmin = MINNONE;        /* Minim: NONE NULL NET DATA; implied by S/ETAG*/
etictr = 0;               /* Number of "NET enabled" tags on stack. */
etisw = 0;                /* 1=tag ended with eti; 0=did not. */
indtdsw = 0;              /* Are we in the DTD? 1=yes; 0=no. */
mslevel = 0;              /* Nesting level of marked sections. */
msplevel = 0;             /* Nested MS levels subject to special parse. */
prologsw = 1;             /* 1=in prolog; 0=not. */
pss = 0;                  /* SGMLACT: scbsgml stack level. */
sgmlsw = 0;               /* SGML declaration found: 1=yes; 0=no. */
stagmin = MINNONE;        /* Minimization: NONE, NULL tag, implied by STAG*/
tagctr = 0;               /* Tag source chars read. */
ts = -1;                  /* Index of current tag in stack. */

docetd = 0;       /* The etd for the document as a whole. */
etagreal = 0;     /* Actual or dummy etd that implied this tag. */
newetd = 0;       /* The etd for a start- or end-tag recognized. */
nextetd = 0;      /* ETD that must come next (only one choice). */
stagreal = 0;     /* Actual or dummy etd that implied this tag. */
conpcb = 0;     /* Current PCB for content parse. */
data = 0;               /* Pointer to returned data in buffer. */
mdname = 0;             /* Name of current markup declaration. */
subdcl = 0;             /* Subject of markup declaration (e.g., GI). */
Tstart = 0;               /* Save starting token for AND group testing. */
conradn = 0;              /* 1=CONREF attribute in list (0=no). */
datalen = 0;              /* Length of returned data in buffer. */
entlen = 0;               /* Length of TAG or EXTERNAL entity text. */
idadn = 0;                /* Number of ID attribute (0 if none). */
noteadn = 0;              /* Number of NOTATION attribute (0 if none). */
reqadn = 0;               /* Num of atts with REQUIRED default (0=none). */

/* Variable arrays and structures.
*/
al = 0;            /* Current attribute list work area. */
nmgrp = 0;	      /* Element name group */
nnmgrp = 0;	      /* Notation name group */
scbsgml = 0;  /* SGMLACT: return action state stack. */
dtype = 0;              /* Document type name. */
}
